package org.netkernel.layer0.meta.impl;

import java.util.List;

import javax.xml.parsers.ParserConfigurationException;

import org.netkernel.container.IMessages;
import org.netkernel.layer0.bnf.BNFGrammarFactory;
import org.netkernel.layer0.bnf.GrammarParseException;
import org.netkernel.layer0.bnf.IGrammarGroup;
import org.netkernel.layer0.meta.IEndpointArgumentMeta;
import org.netkernel.layer0.urii.ParsedIdentifierImpl;
import org.netkernel.layer0.util.FastSchematron;
import org.netkernel.layer0.util.PairList;
import org.netkernel.layer0.util.Utils;
import org.netkernel.layer0.util.XMLReadable;
import org.netkernel.urii.INetKernelException;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

public class ActiveGrammarParser implements IGrammarParser
{
	private static final PairList VALIDATE_ACTIVE_GRAMMAR = new PairList(5);
	static
	{	VALIDATE_ACTIVE_GRAMMAR.put("*[name()!='identifier' and name()!='argument' and name()!='varargs']", "MSG_GRAMMAR_ACTIVE_ROOT_TAGS");
		VALIDATE_ACTIVE_GRAMMAR.put("../active[count(identifier)=0]", "MSG_GRAMMAR_ACTIVE_NO_IDENTIFIER");
		VALIDATE_ACTIVE_GRAMMAR.put("*/*", "MSG_GRAMMAR_ACTIVE_NO_DEEPER_TAGS");
		//VALIDATE_ACTIVE_GRAMMAR.put("argument/attribute::*[name()!='name' and name()!='desc' and name()!='max' and name()!='min' and name()!='regex' and name()!='regexraw']", "MSG_GRAMMAR_ACTIVE_ARG_ATTS");
		VALIDATE_ACTIVE_GRAMMAR.put("argument/attribute::*[name()!='name' and name()!='desc' and name()!='max' and name()!='min' ]", "MSG_GRAMMAR_ACTIVE_ARG_ATTS");
		VALIDATE_ACTIVE_GRAMMAR.put("type/attribute::*", "MSG_GRAMMAR_ACTIVE_NO_DEEPER_ATTRIBUTES");
		VALIDATE_ACTIVE_GRAMMAR.put("varargs/attribute::*[name()!='name' and name()!='desc']", "MSG_GRAMMAR_ACTIVE_VARARG_ATTRIBUTES");	
	};

	public void parseGrammar(Element aGrammarDefn, EndpointMetaBuilder aBuilder, IMessages aMessages) 	throws ParserConfigurationException, GrammarParseException, INetKernelException
	{
		PairList validation=FastSchematron.validate(aGrammarDefn, VALIDATE_ACTIVE_GRAMMAR);
		if (validation.size()>0)
		{	String m = FastSchematron.formatValidationError(validation,aMessages);			
			throw Utils.createFormattedException("EX_STD_MAPPER_CONFIG", "MSG_RAW", aMessages, m);
		}
		
		XMLReadable r = new XMLReadable(aGrammarDefn);
	
		IGrammarGroup grammarRoot = BNFGrammarFactory.createGroup(null);
		List<String> identifiers=r.getTexts("identifier", true);
		int is=identifiers.size();
		
		//parse identifiers
		String[] schemes = new String[is];
		String[] types = new String[is];
		for (int i=0; i<is; i++)
		{	String identifier=identifiers.get(i);
			int j=identifier.indexOf(':');
			if (j>=0)
			{	schemes[i]=identifier.substring(0,j);
				types[i]=identifier.substring(j+1);
			}
			else
			{	schemes[i]="";
				types[i]=identifier;
			}
		}
		//check for scheme homogeneity
		boolean schemeHomogeneity=true;
		String firstScheme=schemes[0];
		for (int i=1; i<is; i++)
		{	if (!schemes[i].equals(firstScheme))
			{	schemeHomogeneity=false;
				break;
			}
		}
		
		//build appropriate structure
		if (schemeHomogeneity)
		{	IGrammarGroup scheme = BNFGrammarFactory.createGroup(grammarRoot,ParsedIdentifierImpl.ARG_ACTIVE_SCHEME,null);
			BNFGrammarFactory.createText(scheme, firstScheme);
			BNFGrammarFactory.createText(grammarRoot, ":");
			if (is>1)
			{	IGrammarGroup typeGroup = BNFGrammarFactory.createChoiceGroup(grammarRoot, ParsedIdentifierImpl.ARG_ACTIVE_TYPE,null);
				for (String typeValue: types)
				{	BNFGrammarFactory.createText(typeGroup, typeValue);
				}
			}
			else
			{	IGrammarGroup typeGroup=BNFGrammarFactory.createGroup(grammarRoot, ParsedIdentifierImpl.ARG_ACTIVE_TYPE,null);
				BNFGrammarFactory.createText(typeGroup, types[0]);
			}
		}
		else
		{	IGrammarGroup choices = BNFGrammarFactory.createChoiceGroup(grammarRoot);
			for (int i=0; i<is; i++)
			{	IGrammarGroup choice = BNFGrammarFactory.createGroup(choices);
				IGrammarGroup scheme = BNFGrammarFactory.createGroup(choice,ParsedIdentifierImpl.ARG_ACTIVE_SCHEME,null);
				BNFGrammarFactory.createText(scheme, schemes[i]);
				IGrammarGroup type = BNFGrammarFactory.createGroup(choice,ParsedIdentifierImpl.ARG_ACTIVE_TYPE,null);
				BNFGrammarFactory.createText(type, types[i]);
			}
		}
		
		List arguments = r.getNodes("argument");
		List<Node> varargs=r.getNodes("varargs");
		boolean hasVarargs=varargs.size()>0;
		IGrammarGroup interleave=null;
		if (arguments.size()>0 || hasVarargs)
		{	interleave=BNFGrammarFactory.createInterleave(grammarRoot);
		}
		
		for (int i=0; i<arguments.size(); i++)
		{	XMLReadable arg=new XMLReadable((Node)arguments.get(i));
			int min=1;
			int max=1;
			String t=arg.getText("@min");
			if (t.length()>0)
			{	min=Integer.parseInt(t);
			}
			t=arg.getText("@max");
			if (t.length()>0)
			{	if (t.equals("*"))
				{	max = Integer.MAX_VALUE;
				}
				else
				{	max=Integer.parseInt(t);
				}
			}
			String name=arg.getText("@name");
			String desc=arg.getText("@desc");
			if (desc.length()==0) desc=null;
			
			MetadataUtils.createArgument(interleave, name, min, max);
			
			/********
			Unused functionality to support @regex and @regexraw attributes to offer regex matching on the argument.
			This is not used in order to avoid accidental affects of over-constraining the active: grammar such as breaking pbv: arguments.
			*********
			String regex=arg.getText("@regex");
			if (regex.length()==0) regex=null;
			String regexraw=arg.getText("@regexraw");
			if (regexraw.length()==0) regexraw=null;
			boolean raw=regex==null && regexraw!=null;
			if(raw)
			{	regex=regexraw;
			}
			MetadataUtils.createArgument(interleave, name, min, max, regex, raw);  //See commented method in MetadataUtils
			***********/
					
			//IEndpointArgumentMeta argMeta = new EndpointArgumentMetaImpl(name,desc,null,null,null,true,null);
			IEndpointArgumentMeta argMeta = new SourcedArgumentMetaImpl(name, desc, null, null);
			aBuilder.addArgument(argMeta);			
		}
		
		if (hasVarargs)
		{	MetadataUtils.createVarargs(interleave);
			Element varargNode=(Element)varargs.get(0);
			String varargsName=varargNode.getAttribute("name");
			if (varargsName.length()==0) varargsName="varargs";
			String varargsDesc=varargNode.getAttribute("desc");
			if (varargsDesc.length()==0) varargsDesc=null;
			SourcedArgumentMetaImpl argMeta = new SourcedArgumentMetaImpl("varargs", varargsDesc, varargsName, null);
			aBuilder.addArgument(argMeta);	
			
		}
		aBuilder.setGrammarRoot(grammarRoot);				
	}

}
