/******************************************************************************
 (c) Copyright 2002-2006, 1060 Research Ltd

 This Software is licensed to You, the licensee, for use under the terms of
 the 1060 Public License v1.0. Please read and agree to the 1060 Public
 License v1.0 [www.1060research.com/license] before using or redistributing
 this software.

 In summary the 1060 Public license has the following conditions.
 A. You may use the Software free of charge provided you agree to the terms
 laid out in the 1060 Public License v1.0
 B. You are only permitted to use the Software with components or applications
 that provide you with OSI Certified Open Source Code [www.opensource.org], or
 for which licensing has been approved by 1060 Research Limited.
 You may write your own software for execution by this Software provided any
 distribution of your software with this Software complies with terms set out
 in section 2 of the 1060 Public License v1.0
 C. You may redistribute the Software provided you comply with the terms of
 the 1060 Public License v1.0 and that no warranty is implied or given.
 D. If you find you are unable to comply with this license you may seek to
 obtain an alternative license from 1060 Research Limited by contacting
 license@1060research.com or by visiting www.1060research.com

 NO WARRANTY:  THIS SOFTWARE IS NOT COVERED BY ANY WARRANTY. SEE 1060 PUBLIC
 LICENSE V1.0 FOR DETAILS

 THIS COPYRIGHT NOTICE IS *NOT* THE 1060 PUBLIC LICENSE v1.0. PLEASE READ
 THE DISTRIBUTED 1060_Public_License.txt OR www.1060research.com/license

 File:          $RCSfile$
 Version:       $Name$ $Revision$
 Last Modified: $Date$
 *****************************************************************************/
package org.netkernel.layer0.tools;

import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.util.*;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.atomic.AtomicLong;
import java.util.regex.*;
import org.netkernel.container.IKernel;
import org.netkernel.container.config.IConfiguration;
import org.netkernel.layer0.boot.BootOrderOptimizer;
import org.netkernel.request.IRequest;
import org.netkernel.request.IResponse;
import org.netkernel.request.IResponseMeta;
import org.netkernel.request.impl.RequestFactory;
import org.netkernel.scheduler.IInternalResponseMeta;
import org.netkernel.urii.IEndpoint;
import org.netkernel.urii.ISpace;

/** Registered as a callback on the kernel this class captures data on the execution of endpoints
 * @author Tony Butterfield
 */
public class EndpointMonitor implements IEndpointMonitor
{
	private Map mRootRequestToTrace;
	private boolean mTraceEnabled=false;
	private Matcher mTraceRootIdentifierMatch;
	
	protected IKernel mKernel;
	
	private AtomicLong mCostFromCache=new AtomicLong();
	private AtomicLong mCostFromCompute=new AtomicLong();
	private AtomicInteger mCostCount = new AtomicInteger();
	
	public static final IEndpoint sInternalDependency= new IEndpoint()
	{
		public void onAsyncRequest(IRequest aRequest)
		{
		}

		public void onCommissionEndpoint(IKernel aKernel, ISpace aSpace) throws Exception
		{
		}

		public void onDecommissionEndpoint() throws Exception
		{
		}

		public boolean isThreadSafe()
		{	return true;
		}
	};
	
	public static class RequestTrace
	{
		private RequestCapture mRootCapture;
		private Map mRequestToCapture;
		private long mStart;
		private long mEnd;
		private Object mTransportInstance;
		private ISpace mTransportContext;
		
		public RequestTrace(Object aTransportInstance, ISpace aTransportContext, IRequest aRootRequest)
		{	
			mRootCapture=new RequestCapture(aRootRequest,null);
			mRequestToCapture=new IdentityHashMap();
			mStart = mEnd = System.currentTimeMillis();
			mTransportInstance=aTransportInstance;
			mTransportContext=aTransportContext;
		}
		
		public void setTransport(Object aTransportInstance)
		{	mTransportInstance=aTransportInstance;
		}
		
		private RequestCapture forRequest(IRequest aRequest)
		{	RequestCapture result;
			if (aRequest==null)
			{	result=mRootCapture;
			}
			else
			{	result=(RequestCapture)mRequestToCapture.get(aRequest);
			}
			return result;
		}
		
		public void endpointInvoke(IRequest aRequestIn, IRequest aRequestOut, IRequest aParent)
		{	RequestCapture rc = new RequestCapture(aRequestIn,aRequestOut);
			RequestCapture parentRc = forRequest(aParent);
			if (parentRc!=null)
			{	parentRc.addSubRequest(rc);
			}
			mRequestToCapture.put(aRequestIn,rc);
		}
		
		public void endpointResponse(IRequest aRequest, IEndpoint aEndpoint, long aDuration, IResponse aResponse)
		{	RequestCapture rc = forRequest(aRequest);
			if (rc!=null)
			{	rc.requestResponse(aDuration,aEndpoint,aResponse);
			}
			if (aRequest.getParent()==null)
			{	mEnd=System.currentTimeMillis();
			}
		}
		
		public void resolutionFailure(IRequest aRequestIn, IRequest aParent, IResponse aResponse)
		{	RequestCapture rc = new RequestCapture(aRequestIn,null);
			rc.mResponse=aResponse;
			RequestCapture parentRc = forRequest(aParent);
			if (parentRc!=null)
			{	parentRc.addSubRequest(rc);
			}
			mRequestToCapture.put(aRequestIn,rc);
		}
		
		public RequestCapture getRoot()
		{	RequestCapture result=null;
			List subReqs=mRootCapture.getSubRequests();
			if (subReqs.size()>0)
			{	result = (RequestCapture)subReqs.get(0);
			}
			return result;
		}
		
		public IRequest getRootRequest()
		{	return mRootCapture.mRequestIn;
		}
		
		public long getStart()
		{	return mStart;
		}
		public long getEnd()
		{	return mEnd;
		}
		
		public Object getTransport()
		{	return mTransportInstance;
		}
		
		public ISpace getTransportContext()
		{	return mTransportContext;
		}
		
	}
	
	public static class RequestCapture
	{	private List mSubRequests;
		private long mDuration;
		private IRequest mRequestIn;
		private IRequest mRequestOut;
		private IEndpoint mEndpoint;
		private IResponse mResponse;
		private RequestCapture mParent;
		private long mTime;
		private boolean mExpired;
		
		private static final IResponse[] sNoAdditionalDependencies=new IResponse[0];
		
		
		public RequestCapture(IRequest aRequestIn, IRequest aRequestOut)
		{	mRequestIn=aRequestIn;
			mRequestOut=aRequestOut;
			mSubRequests=new ArrayList();
			if (aRequestOut!=null)
			{	mTime=System.currentTimeMillis();
			}
		}
		
		private void setParent(RequestCapture aCapture)
		{	mParent=aCapture;
		}
		
		public void requestResponse(long aDuration, IEndpoint aEndpoint, IResponse aResponse)
		{	mDuration=aDuration;
			mEndpoint=aEndpoint;
			mResponse=aResponse;
			long now=System.currentTimeMillis();
			mTime=now-mTime;
			mExpired=mResponse.getMeta().getExpiry().isExpired(now);
			List<IResponse> ad=(List<IResponse>)mResponse.getMeta().getUserMetaData().getValue(IResponseMeta.KEY_ADDITIONAL_DEPENDENCIES);
			if (ad!=null && ad.size()>0)
			{	for (int i=0; i<ad.size(); i++)
				{	IResponse r=ad.get(i);
					RequestCapture rc=new RequestCapture(r.getRequest(), r.getRequest());
					rc.requestResponse(0, sInternalDependency, r);
					addSubRequest(rc);
				}
				
			}
		}
		
		public void addSubRequest(RequestCapture aSubRequest)
		{	aSubRequest.setParent(this);
			synchronized(mSubRequests)
			{	mSubRequests.add(aSubRequest);
			}
		}
		
		public List getSubRequests()
		{	return mSubRequests;
		}
		
		public IEndpoint getEndpoint()
		{	return mEndpoint;
		}
		
		public long getDuration()
		{	return mDuration;
		}
		
		public long getElapsed()
		{	return mTime;
		}
		
		public IRequest getRequest()
		{	return mRequestOut;
		}
		
		public IRequest getRequestIn()
		{	return mRequestIn;
		}
		
		public IResponse getResponse()
		{	return mResponse;
		}
		
		public boolean isExpired()
		{	return mExpired;
		}
		
		public RequestCapture getParent()
		{	return mParent;
		}		
	}
	
	public EndpointMonitor(IKernel aKernel)
	{	
		mRootRequestToTrace = new IdentityHashMap();
		mKernel=aKernel;
	}
	
	private RequestTrace getRequestTrace(IRequest aRootRequest)
	{	RequestTrace trace = (RequestTrace)mRootRequestToTrace.get(aRootRequest);
		return trace;
	}
	
	private boolean matchedRequest(IRequest aRootRequest)
	{	boolean result=true;
		String identifierString = aRootRequest.getIdentifier().toString();
		if (mTraceRootIdentifierMatch!=null)
		{	synchronized(mTraceRootIdentifierMatch)
			{	mTraceRootIdentifierMatch.reset(identifierString);
				result=mTraceRootIdentifierMatch.matches();
			}
		}
		return result;
	}
	
	public void externalRequest(IRequest aRootRequest, Object aTransportInstance, ISpace aTransportSpace)
	{	if (mTraceEnabled && matchedRequest(aRootRequest))
		{	RequestTrace trace = new RequestTrace(aTransportInstance, aTransportSpace,aRootRequest); 
			mRootRequestToTrace.put(aRootRequest,trace);
		}
	}

	public void cacheHit(IRequest aRootRequest, IRequest aParentRequest, IRequest aRequest, IResponse aResponse)
	{	if (mTraceEnabled && matchedRequest(aRootRequest))
		{	RequestTrace trace = getRequestTrace(aRootRequest);
			if (trace!=null)
			{	trace.endpointInvoke(aRequest,aRequest,aParentRequest);
				trace.endpointResponse(aRequest, null,0,aResponse);
			}
		}
	
		if (BootOrderOptimizer.sInstance.isActive() && aRequest.getVerb()==IRequest.VERB_META && aRequest.getIdentifier().toString().length()==0)
		{	BootOrderOptimizer.sInstance.enterSpace(aRequest.getRequestScope().getSpace(),true);
			BootOrderOptimizer.sInstance.exitSpace(aRequest.getRequestScope().getSpace());
		}
		/*
		if (aResponse.getMeta() instanceof IInternalResponseMeta)
		{	int cost=((IInternalResponseMeta)aResponse.getMeta()).getTotalCost();
			mCostFromCache.addAndGet(cost);
			if (mCostCount.incrementAndGet()==COST_PERIOD)
			{	processCosts();
			}
		}
		*/	
	}
	
	public void resolutionFailure(IRequest aRootRequest, IRequest aParentRequest, IRequest aRequest, IResponse aResponse)
	{	if (mTraceEnabled && matchedRequest(aRootRequest))
		{	RequestTrace trace = getRequestTrace(aRootRequest);
			if (trace!=null)
			{	trace.resolutionFailure(aRequest,aParentRequest,aResponse);
			}
		}
	}
	
	public void endpointInvoke(IRequest aRootRequest, IRequest aParentRequest, IRequest aRequestIn, IRequest aRequestOut, IEndpoint aEndpoint )
	{	if (mTraceEnabled && matchedRequest(aRootRequest))
		{	RequestTrace trace = getRequestTrace(aRootRequest);
			if (trace!=null)
			{	//System.out.println("endpointInvoke "+trace.getTransport());
				trace.endpointInvoke(aRequestIn,aRequestOut,aParentRequest);
			}
		}
	
		if (BootOrderOptimizer.sInstance.isActive() && aRequestIn.getVerb()==IRequest.VERB_META && aRequestIn.getIdentifier().toString().length()==0)
		{	BootOrderOptimizer.sInstance.enterSpace(aRequestIn.getRequestScope().getSpace(),false);
		}
	
	
	}

	public void endpointResponse(IRequest aRootRequest, IRequest aRequestIn, IEndpoint aEndpoint, long aDuration, IResponse aResponse)
	{
		if (mTraceEnabled && matchedRequest(aRootRequest))
		{
			//System.out.println("endpointResponse "+aRootRequest.toString());
			RequestTrace trace = getRequestTrace(aRootRequest);
			if (trace!=null)
			{	trace.endpointResponse(aRequestIn, aEndpoint, aDuration, aResponse);
			}
		}
		
		if (BootOrderOptimizer.sInstance.isActive() && aRequestIn.getVerb()==IRequest.VERB_META && aRequestIn.getIdentifier().toString().length()==0)
		{	BootOrderOptimizer.sInstance.exitSpace(aRequestIn.getRequestScope().getSpace());
		}
		/*
		if (aResponse.getMeta() instanceof IInternalResponseMeta)
		{	int cost=((IInternalResponseMeta)aResponse.getMeta()).getLocalCost();
			mCostFromCompute.addAndGet(cost);
			if (mCostCount.incrementAndGet()==COST_PERIOD)
			{	processCosts();
			}
		}
		*/
		
	}
	
	private final static int COST_PERIOD = 512;
	private double mF=0.0;
	NumberFormat mNF=new DecimalFormat("#.000");
	
	private void processCosts()
	{	
		mCostCount.set(0);
		long costCompute=mCostFromCompute.getAndSet(0);
		long costCache=mCostFromCache.getAndSet(0);
		double f=((double)costCache)/((double)(costCompute+costCache));
		if (mF==0.0)
		{	mF=f;
		}
		else
		{	mF=mF*0.9+f*0.1;
		}
		System.out.println("UNITY "+mNF.format(f)+" "+mNF.format(mF));
	}
	
	public synchronized void clear()
	{	
	}
	
	//trace stuff
	public void enableTrace(boolean aEnabled)
	{	mTraceEnabled=aEnabled;
	}
	public boolean isTraceEnabled()
	{	return mTraceEnabled;
	}
	public String getRootTraceMatch()
	{	return (mTraceRootIdentifierMatch!=null)?mTraceRootIdentifierMatch.pattern().pattern():"";
	}
	public void setRootTraceMatch(String aPattern) throws PatternSyntaxException
	{	if (aPattern!=null && aPattern.length()>0)
		{	mTraceRootIdentifierMatch=Pattern.compile(aPattern).matcher("");
		}
		else
		{	mTraceRootIdentifierMatch=null;
		}
	}
	public void clearCapturedTraces()
	{	mRootRequestToTrace.clear();
	}
	
	public void removeCapturedTrace(IRequest aRootRequest)
	{	mRootRequestToTrace.remove(aRootRequest);
	}
	
	public RequestTrace[] getCapturedTraces()
	{	
		Collection c=mRootRequestToTrace.values();
		int s=0;
		for (Iterator i=c.iterator(); i.hasNext(); )
		{	RequestTrace trace=(RequestTrace)i.next();
			if (trace.getRoot()!=null)
			{	s++;
			}
		}
		RequestTrace[] result = new RequestTrace[s];
		int j=0;
		for (Iterator i=c.iterator(); i.hasNext(); )
		{	RequestTrace trace=(RequestTrace)i.next();
			if (trace.getRoot()!=null)
			{	result[j++]=trace;
			}
		}
		return result;
	}

	public void configurationChanged(IConfiguration aConfiguration)
	{	// nothing to do yet
	}
}