package org.netkernel.layer0.util;

import java.io.*;
import java.net.*;
import java.util.*;

import javax.xml.xpath.XPathExpressionException;

import org.netkernel.container.IMessages;
import org.netkernel.layer0.nkf.INKFLocale;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;


public class FastSchematron
{	
	private static Map<PairList, Integer> sValidationHash= new HashMap<PairList,Integer>();
	private static Map<Integer,Boolean> sValidationCache = new HashMap<Integer,Boolean>();
	private static URL sPersistenceLocation;
	private static boolean sUsePersistenceCache;
	private static boolean sSaved=true;
	
	public static int fragmentHash(Element aElement)
	{
		int hash=aElement.getNodeName().hashCode();
		
		NamedNodeMap atts=aElement.getAttributes();
		for (int i=atts.getLength()-1; i>=0; i--)
		{	Node att=atts.item(i);
			hash^=att.getNodeName().hashCode();
			hash^=att.getNodeValue().hashCode();
		}

		Node n=aElement.getFirstChild();
		while (n!=null)
		{	switch (n.getNodeType())
			{	case Node.TEXT_NODE:
					hash^=n.getNodeValue().hashCode();
					break;
				case Node.ELEMENT_NODE:
					hash^=fragmentHash((Element)n);
					break;
			}
			n=n.getNextSibling();
		}
		return hash;
	}
	
	public static int pairListHash(PairList aPairList)
	{	int hash=0;
		for (int i=aPairList.size()-1; i>=0; i--)
		{	hash^=aPairList.getValue1(i).hashCode();
			hash^=aPairList.getValue2(i).hashCode();
		}
		return hash;
	}
	
	public static void loadValidationCache()
	{	
		//ensure xpath if initialised by main thread even if we don't need it 
		XPath.init();
		sUsePersistenceCache=false;
		
		sPersistenceLocation= FastSchematron.class.getClassLoader().getResource("etc/validationCache");
		if (sPersistenceLocation==null)
		{	URL dir=FastSchematron.class.getClassLoader().getResource("etc/");
			if (dir!=null && dir.getProtocol().equals("file"))
			{	//work around no distro created file
				try
				{	File dirFile=new File(dir.toURI());
					new File(dirFile,"validationCache").createNewFile();
					sPersistenceLocation= FastSchematron.class.getClassLoader().getResource("etc/validationCache");
				}
				catch (Exception e) {;}
			}
		}
				
		if (sPersistenceLocation!=null)
		{	try
			{	InputStream is=sPersistenceLocation.openConnection().getInputStream();
				BufferedReader br = new BufferedReader(new InputStreamReader(is));
				String hashString;
				while ( (hashString=br.readLine())!=null )
				{	int hash=Integer.parseInt(hashString);
					sValidationCache.put(hash,Boolean.FALSE);
				}
				br.close();
			}
			catch (IOException e)
			{	e.printStackTrace();	
			}
			
			sUsePersistenceCache=sValidationCache.size()>0 || sPersistenceLocation.getProtocol().equals("file");
			sSaved=true;
		}
	}
	
	public static void saveValidationCache()
	{	if (sPersistenceLocation!=null && sPersistenceLocation.getProtocol().equals("file") && !sSaved)
		{	boolean useCacheBackup=sUsePersistenceCache;
			sUsePersistenceCache=false;
			try
			{	File persistenceFile = new File(sPersistenceLocation.toURI());
				BufferedWriter bw = new BufferedWriter(new FileWriter(persistenceFile));
				for (Map.Entry<Integer,Boolean> entry : sValidationCache.entrySet())
				{	if (entry.getValue().booleanValue())
					{	bw.write(entry.getKey().toString());
						bw.newLine();
					}
				}
				bw.flush();
				bw.close();
				sSaved=true;
			}
			catch (Exception e)
			{	//e.printStackTrace();	
			}
			sUsePersistenceCache=useCacheBackup;
		}
	}
	
	/** Validate the given document relative to the given element by applying the rules. If
	 * any rules return false then the corresponding errors are added to the return list.
	 * @param aRoot the root of validation to which all relative rules will be evaluated against
	 * @param aRulesToErrors a pairlist containing xpath_string -> error_object pairs. The objects
	 * can be any application specific identifier or value.  
	 * @return a pairlist containing zero or more node -> error_object pairs
	 * @exception IllegalArgumentException if there are any problems evaluation the xpaths
	 */
	public static PairList validate(Element aRoot, PairList aRulesToErrors) throws IllegalArgumentException
	{	
		Integer combinedHash=null;
		if (sUsePersistenceCache)
		{	Integer validationHash=sValidationHash.get(aRulesToErrors);
			if (validationHash==null)
			{	validationHash=pairListHash(aRulesToErrors);
				sValidationHash.put(aRulesToErrors, validationHash);
			}
			
			int documentHash = fragmentHash(aRoot);
			
			combinedHash=documentHash^validationHash;
			Boolean found = sValidationCache.get(combinedHash);
			if (found!=null)
			{	if (!found.booleanValue())
				{	sValidationCache.put(combinedHash, Boolean.TRUE);
				}
				return PairList.EMPTY;
			}
		}
		
		PairList result = PairList.EMPTY;
		try
		{	int size=aRulesToErrors.size();
			if (size>0)
			{	result = new PairList(aRulesToErrors.size());
				for (int i=0; i<size; i++)
				{	String xpath=(String)aRulesToErrors.getValue1(i);
					List failedNodes = XPath.eval(xpath,aRoot);
					for(int j=0; j<failedNodes.size(); j++)
					{	result.put(failedNodes.get(j),aRulesToErrors.getValue2(i));
					}
				}
			}
		}
		catch (XPathExpressionException e)
		{	throw new IllegalArgumentException(e);
		}
		
		if (combinedHash!=null && result.size()==0 )
		{	sValidationCache.put(combinedHash,Boolean.TRUE);
			sSaved=false;
		}
		
		return result;
	}
	
	public static String formatValidationError(PairList aValidation, IMessages aMessages)
	{	StringBuilder sb = new StringBuilder(aValidation.size()*32);
		for (int i=0; i<aValidation.size(); i++)
		{	if (i>0)
			{	sb.append(", ");
			}
			String errorKey = (String)aValidation.getValue2(i);
			Node errorNode = (Node)aValidation.getValue1(i);
			String m = aMessages.format(errorKey, XMLUtils.getPathFor(errorNode));
			sb.append(m);
		}
		return sb.toString();
	}
	
	public static String formatValidationError(PairList aValidation, INKFLocale aLocale)
	{	StringBuilder sb = new StringBuilder(aValidation.size()*32);
		for (int i=0; i<aValidation.size(); i++)
		{	if (i>0)
			{	sb.append(", ");
			}
			String errorKey = (String)aValidation.getValue2(i);
			Node errorNode = (Node)aValidation.getValue1(i);
			String m = aLocale.formatMessage(errorKey, XMLUtils.getPathFor(errorNode));
			sb.append(m);
		}
		return sb.toString();
	}

}